@@ -5,7 +5,14 @@ import struct
 import time
 import ssl
 from enum import IntEnum
-from .constants import TCP_DATA_UNSECURED, UDP_DISCOVERY, A_PROCESSING_TIME, LINK_LOCAL_MULTICAST_ADDRESS
+from typing import Union
+from .constants import (
+    A_DOIP_CTRL,
+    TCP_DATA_UNSECURED,
+    UDP_DISCOVERY,
+    A_PROCESSING_TIME,
+    LINK_LOCAL_MULTICAST_ADDRESS,
+)
 from .messages import *
 
 logger = logging.getLogger("doipclient")
@@ -68,7 +75,7 @@ class Parser:
                 self.payload_type = self.rx_buffer.pop(0) << 8
                 self.payload_type |= self.rx_buffer.pop(0)
                 self._state = Parser.ParserState.READ_PAYLOAD_SIZE
-        
+
         if self._state == Parser.ParserState.READ_PAYLOAD_SIZE:
             if len(self.rx_buffer) >= 4:
                 self.payload_size = self.rx_buffer.pop(0) << 24
@@ -76,7 +83,7 @@ class Parser:
                 self.payload_size |= self.rx_buffer.pop(0) << 8
                 self.payload_size |= self.rx_buffer.pop(0)
                 self._state = Parser.ParserState.READ_PAYLOAD
-        
+
         if self._state == Parser.ParserState.READ_PAYLOAD:
             remaining_bytes = self.payload_size - len(self.payload)
             self.payload += self.rx_buffer[:remaining_bytes]
@@ -110,11 +117,12 @@ class DoIPClient:
 
     :param ecu_ip_address: This is the IP address of the target ECU. This should be a string representing an IPv4
         address like "192.168.1.1" or an IPv6 address like "2001:db8::". Like the logical_address, if you don't know the
-        value for your ECU, utilize the await_vehicle_announcement() method.
+        value for your ECU, utilize the get_entity() or await_vehicle_announcement() method.
     :type ecu_ip_address: str
     :param ecu_logical_address: The logical address of the target ECU. This should be an integer. According to the
         specification, the correct range is 0x0001 to 0x0DFF ("VM specific"). If you don't know the logical address,
-        use the await_vehicle_announcement() method and power cycle the ECU - it should identify itself on bootup.
+        either use the get_entity() method OR the await_vehicle_announcement() method and power
+        cycle the ECU - it should identify itself on bootup.
     :type ecu_logical_address: int
     :param tcp_port: The destination TCP port for DoIP data communication. By default this is 13400 for unsecure and
         3496 when using TLS.
@@ -134,12 +142,13 @@ class DoIPClient:
         Useful if you have multiple network adapters. Can be an IPv4 or IPv6 address just like `ecu_ip_address`, though
         the type should match.
     :type client_ip_address: str, optional
-    :param use_secure: Enables TLS if True. Untested. Should be combined with changing tcp_port to 3496.
-    :type use_secure: bool
-    :param log_level: Logging level
-    :type log_level: int
+    :param use_secure: Enables TLS. If set to True, a default SSL context is used. For more control, a preconfigured
+        SSL context can be passed directly. Untested. Should be combined with changing tcp_port to 3496.
+    :type use_secure: Union[bool,ssl.SSLContext]
     :param auto_reconnect_tcp: Attempt to automatically reconnect TCP sockets that were closed by peer
     :type auto_reconnect_tcp: bool
+    :param vm_specific: Optional 4 byte long int
+    :type vm_specific: int, optional
 
     :raises ConnectionRefusedError: If the activation request fails
     :raises ValueError: If the IPAddress is neither an IPv4 nor an IPv6 address
@@ -157,6 +166,7 @@ class DoIPClient:
         client_ip_address=None,
         use_secure=False,
         auto_reconnect_tcp=False,
+        vm_specific=None,
     ):
         self._ecu_logical_address = ecu_logical_address
         self._client_logical_address = client_logical_address
@@ -171,6 +181,7 @@ class DoIPClient:
         self._protocol_version = protocol_version
         self._auto_reconnect_tcp = auto_reconnect_tcp
         self._tcp_close_detected = False
+        self.vm_specific = vm_specific
 
         # Check the ECU IP type to determine socket family
         # Will raise ValueError if neither a valid IPv4, nor IPv6 address
@@ -198,9 +209,65 @@ class DoIPClient:
     def __exit__(self, type, value, traceback):
         self.close()
 
+    @staticmethod
+    def _create_udp_socket(
+        ipv6=False, udp_port=UDP_DISCOVERY, timeout=None, source_interface=None
+    ):
+        if ipv6:
+            sock = socket.socket(socket.AF_INET6, socket.SOCK_DGRAM)
+
+            # IPv6 version always uses link-local scope multicast address (FF02 16 ::1)
+            sock.bind((LINK_LOCAL_MULTICAST_ADDRESS, udp_port))
+
+            if source_interface is None:
+                # 0 is the "default multicast interface" which is unlikely to be correct, but it will do
+                interface_index = 0
+            else:
+                interface_index = socket.if_nametoindex(source_interface)
+
+            # Join the group so that packets are delivered
+            mc_addr = ipaddress.IPv6Address(LINK_LOCAL_MULTICAST_ADDRESS)
+            join_data = struct.pack("16sI", mc_addr.packed, interface_index)
+            # IPV6_JOIN_GROUP is also known as IPV6_ADD_MEMBERSHIP, though older Python for Windows doesn't have it
+            # IPPROTO_IPV6 may be missing in older Windows builds
+            try:
+                from socket import IPPROTO_IPV6
+            except ImportError:
+                IPPROTO_IPV6 = 41
+            sock.setsockopt(IPPROTO_IPV6, socket.IPV6_JOIN_GROUP, join_data)
+        else:
+            sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
+            # IPv4, use INADDR_ANY to listen to all interfaces for broadcasts (not multicast)
+            sock.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)
+            sock.bind(("", udp_port))
+
+        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
+        if timeout is not None:
+            sock.settimeout(timeout)
+
+        return sock
+
+    @staticmethod
+    def _pack_doip(protocol_version, payload_type, payload_data):
+        data_bytes = struct.pack(
+            "!BBHL",
+            protocol_version,
+            0xFF ^ protocol_version,
+            payload_type,
+            len(payload_data),
+        )
+        data_bytes += payload_data
+
+        return data_bytes
+
     @classmethod
     def await_vehicle_announcement(
-        cls, udp_port=UDP_DISCOVERY, timeout=None, ipv6=False, source_interface=None
+        cls,
+        udp_port=UDP_DISCOVERY,
+        timeout=None,
+        ipv6=False,
+        source_interface=None,
+        sock=None,
     ):
         """Receive Vehicle Announcement Message
 
@@ -215,50 +282,26 @@ class DoIPClient:
         :type timeout: float, optional
         :param ipv6: Bool forcing IPV6 socket instead of IPV4 socket
         :type ipv6: bool, optional
-        :return: IP Address of ECU and VehicleAnnouncementMessage object
-        :rtype: tuple
         :param source_interface: Interface name (like "eth0") to bind to for use with IPv6. Defaults to None which
             will use the default interface (which may not be the one connected to the ECU). Does nothing for IPv4,
             which will bind to all interfaces uses INADDR_ANY.
-        :type source_interface: str, optional 
+        :type source_interface: str, optional
+        :return: IP Address of ECU and VehicleAnnouncementMessage object
+        :rtype: tuple
         :raises TimeoutError: If vehicle announcement not received in time
         """
         start_time = time.time()
 
-        if ipv6:
-            sock = socket.socket(socket.AF_INET6, socket.SOCK_DGRAM)
-
-            # IPv6 version always uses link-local scope multicast address (FF02 16 ::1) 
-            sock.bind((LINK_LOCAL_MULTICAST_ADDRESS, udp_port))
-
-            if source_interface is None:
-                # 0 is the "default multicast interface" which is unlikely to be correct, but it will do
-                interface_index = 0
-            else:
-                interface_index = socket.if_nametoindex(source_interface)
-
-            # Join the group so that packets are delivered
-            mc_addr = ipaddress.IPv6Address(LINK_LOCAL_MULTICAST_ADDRESS)
-            join_data = struct.pack('16sI', mc_addr.packed, interface_index)
-            # IPV6_JOIN_GROUP is also known as IPV6_ADD_MEMBERSHIP, though older Python for Windows doesn't have it
-            # IPPROTO_IPV6 may be missing in older Windows builds
-            try:
-                from socket import IPPROTO_IPV6
-            except ImportError:
-                IPPROTO_IPV6 = 41
-            sock.setsockopt(IPPROTO_IPV6, socket.IPV6_JOIN_GROUP, join_data)
-        else:
-            sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
-            # IPv4, use INADDR_ANY to listen to all interfaces for broadcasts (not multicast)
-            sock.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)
-            sock.bind(("", udp_port))
-        
-        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
-        if timeout is not None:
-            sock.settimeout(timeout)
-
         parser = Parser()
 
+        if not sock:
+            sock = cls._create_udp_socket(
+                ipv6=ipv6,
+                udp_port=udp_port,
+                timeout=timeout,
+                source_interface=source_interface,
+            )
+
         while True:
             remaining = None
             if timeout:
@@ -283,6 +326,54 @@ class DoIPClient:
             if result and type(result) == VehicleIdentificationResponse:
                 return addr, result
 
+    @classmethod
+    def get_entity(
+        cls, ecu_ip_address="255.255.255.255", protocol_version=0x02, eid=None, vin=None
+    ):
+        """Sends a VehicleIdentificationRequest and awaits a VehicleIdentificationResponse from the ECU,
+        either with a specified VIN, EIN, or nothing. Equivalent to the request_vehicle_identification() method
+        but can be called without instantiation.
+
+        :param ecu_ip_address: This is the IP address of the target ECU for unicast. Defaults to broadcast if
+        the address is not known.
+        :type ecu_ip_address: str, optional
+        :param protocol_version: The DoIP protocol version to use for communication. Represents the version of the ISO 13400
+            specification to follow. 0x02 (2012) is probably correct for most ECU's at the time of writing, though technically
+            this implementation is against 0x03 (2019).
+        :type protocol_version: int, optional
+        :param eid: EID of the Vehicle
+        :type eid: bytes, optional
+        :param vin: VIN of the Vehicle
+        :type vin: str, optional
+        :return: The vehicle identification response message
+        :rtype: VehicleIdentificationResponse
+        """
+
+        # UDP_TEST_EQUIPMENT_REQUEST is dynamically assigned using udp_port=0
+        sock = cls._create_udp_socket(udp_port=0, timeout=A_DOIP_CTRL)
+
+        if eid:
+            message = VehicleIdentificationRequestWithEID(eid)
+        elif vin:
+            message = VehicleIdentificationRequestWithVIN(vin)
+        else:
+            message = VehicleIdentificationRequest()
+
+        payload_data = message.pack()
+        payload_type = payload_message_to_type[type(message)]
+
+        data_bytes = cls._pack_doip(protocol_version, payload_type, payload_data)
+        logger.debug(
+            "Sending DoIP Vehicle Identification Request: Type: 0x{:X}, Payload Size: {}, Payload: {}".format(
+                payload_type,
+                len(payload_data),
+                " ".join(f"{byte:02X}" for byte in payload_data),
+            )
+        )
+        sock.sendto(data_bytes, (ecu_ip_address, UDP_DISCOVERY))
+
+        return cls.await_vehicle_announcement(timeout=A_DOIP_CTRL, sock=sock)
+
     def empty_rxqueue(self):
         """Implemented for compatibility with udsoncan library. Nothing useful to be done yet"""
         pass
@@ -377,7 +468,7 @@ class DoIPClient:
                     self._tcp_parser.push_bytes(data)
                 # Subsequent reads, go to 0 timeout
                 self._tcp_sock.settimeout(0)
-        except (BlockingIOError, socket.timeout):
+        except (BlockingIOError, socket.timeout, ssl.SSLError):
             pass
         except (ConnectionResetError, BrokenPipeError):
             logger.debug("TCP Connection broken, attempting to reset")
@@ -407,14 +498,7 @@ class DoIPClient:
 
         retry = self._auto_reconnect_tcp and not disable_retry
 
-        data_bytes = struct.pack(
-            "!BBHL",
-            self._protocol_version,
-            0xFF ^ self._protocol_version,
-            payload_type,
-            len(payload_data),
-        )
-        data_bytes += payload_data
+        data_bytes = self._pack_doip(self._protocol_version, payload_type, payload_data)
         logger.debug(
             "Sending DoIP Message: Type: 0x{:X}, Payload Size: {}, Payload: {}".format(
                 payload_type,
@@ -505,16 +589,23 @@ class DoIPClient:
         :type disable_retry: bool, optional
         :return: The resulting activation response object
         :rtype: RoutingActivationResponse
+        :raises ValueError: vm_specific is invalid or out of range
         """
         message = RoutingActivationRequest(
-            self._client_logical_address, activation_type, vm_specific=vm_specific
+            self._client_logical_address,
+            activation_type,
+            vm_specific=(
+                self._validate_vm_specific_value(vm_specific)
+                if vm_specific
+                else self.vm_specific
+            ),
         )
         self.send_doip_message(message, disable_retry=disable_retry)
         while True:
             result = self.read_doip()
-            if type(result) == RoutingActivationResponse:
+            if isinstance(result, RoutingActivationResponse):
                 return result
-            elif result:
+            if result:
                 logger.warning(
                     "Received unexpected DoIP message type {}. Ignoring".format(
                         type(result)
@@ -522,9 +613,8 @@ class DoIPClient:
                 )
 
     def request_vehicle_identification(self, eid=None, vin=None):
-        """Requests a VehicleIdentificationResponse from the ECU, either with a specified VIN, EIN,
+        """Sends a VehicleIdentificationRequest and awaits a VehicleIdentificationResponse from the ECU, either with a specified VIN, EIN,
         or nothing.
-
         :param eid: EID of the Vehicle
         :type eid: bytes, optional
         :param vin: VIN of the Vehicle
@@ -618,12 +708,27 @@ class DoIPClient:
     def send_diagnostic(self, diagnostic_payload, timeout=A_PROCESSING_TIME):
         """Send a raw diagnostic payload (ie: UDS) to the ECU.
 
+        :param diagnostic_payload: UDS payload to transmit to the ECU
+        :type diagnostic_payload: bytearray
+        :raises IOError: DoIP negative acknowledgement received
+        """
+        self.send_diagnostic_to_address(
+            self._ecu_logical_address, diagnostic_payload, timeout
+        )
+
+    def send_diagnostic_to_address(
+        self, address, diagnostic_payload, timeout=A_PROCESSING_TIME
+    ):
+        """Send a raw diagnostic payload (ie: UDS) to the specified address.
+
+        :param address: The logical address to send the diagnostic payload to
+        :type address: int
         :param diagnostic_payload: UDS payload to transmit to the ECU
         :type diagnostic_payload: bytearray
         :raises IOError: DoIP negative acknowledgement received
         """
         message = DiagnosticMessage(
-            self._client_logical_address, self._ecu_logical_address, diagnostic_payload
+            self._client_logical_address, address, diagnostic_payload
         )
         self.send_doip_message(message)
         start_time = time.time()
@@ -693,7 +798,15 @@ class DoIPClient:
             self._udp_sock.bind((self._client_ip_address, 0))
 
         if self._use_secure:
-            self._tcp_sock = ssl.wrap_socket(self._tcp_sock)
+            if isinstance(self._use_secure, ssl.SSLContext):
+                ssl_context = self._use_secure
+            else:
+                ssl_context = ssl.create_default_context()
+            self._wrap_socket(ssl_context)
+
+    def _wrap_socket(self, ssl_context):
+        """Wrap the underlying socket in a SSL context."""
+        self._tcp_sock = ssl_context.wrap_socket(self._tcp_sock)
 
     def close(self):
         """Close the DoIP client"""
@@ -720,3 +833,39 @@ class DoIPClient:
                 raise ConnectionRefusedError(
                     f"Activation Request failed with code {result.response_code}"
                 )
+
+    @staticmethod
+    def _validate_vm_specific_value(value):
+        """Validate the VM specific value (must be > 0 and <= 0xffffffff) or None.
+        If the conditions are not fulfilled, raises an exception.
+
+        :param value: The value to check.
+        :type value: int, optional
+        :return: The input value if valid.
+        :rtype: int or None
+        :raises ValueError: If the value is invalid or out of range.
+        """
+        if not isinstance(value, int) and value is not None:
+            raise ValueError("Invalid vm_specific type must be int or None")
+        if isinstance(value, int) and (value < 0 or value > 0xFFFFFFFF):
+            raise ValueError("Invalid vm_specific value must be > 0 and <= 0xffffffff")
+        return value
+
+    @property
+    def vm_specific(self):
+        """Get the optional OEM specific field value if set.
+
+        :return: vm_specific value
+        :rtype: int, optional
+        """
+        return self._vm_specific
+
+    @vm_specific.setter
+    def vm_specific(self, value):
+        """Set the optional OEM specific field value. If you do not need to send this item, set it to None.
+
+        :param value: The vm_specific value (must be > 0 and <= 0xffffffff) or None
+        :type value: int, optional
+        :raises ValueError: Value is invalid or out of range
+        """
+        self._vm_specific = self._validate_vm_specific_value(value)
