@@ -1,4 +1,5 @@
 import socket
+import ssl
 import pytest
 import logging
 from doipclient import DoIPClient
@@ -98,6 +99,9 @@ diagnostic_power_mode_response = bytearray(
 diagnostic_request = bytearray(
     [int(x, 16) for x in "02 fd 80 01 00 00 00 07 0e 00 00 01 00 01 02".split(" ")]
 )
+diagnostic_request_to_address = bytearray(
+    [int(x, 16) for x in "02 fd 80 01 00 00 00 07 0e 00 12 34 00 01 02".split(" ")]
+)
 unknown_mercedes_message = bytearray(
     [
         int(x, 16)
@@ -499,6 +503,23 @@ def test_send_diagnostic_negative(mock_socket):
     assert mock_socket.tx_queue[-1] == diagnostic_request
 
 
+def test_send_diagnostic_to_address_positive(mock_socket):
+    sut = DoIPClient(test_ip, test_logical_address)
+    mock_socket.rx_queue.append(diagnostic_positive_response)
+    assert None == sut.send_diagnostic_to_address(0x1234, bytearray([0, 1, 2]))
+    assert mock_socket.tx_queue[-1] == diagnostic_request_to_address
+
+
+def test_send_diagnostic_to_address_negative(mock_socket):
+    sut = DoIPClient(test_ip, test_logical_address)
+    mock_socket.rx_queue.append(diagnostic_negative_response)
+    with pytest.raises(
+        IOError, match=r"Diagnostic request rejected with negative acknowledge code"
+    ):
+        result = sut.send_diagnostic_to_address(0x1234, bytearray([0, 1, 2]))
+    assert mock_socket.tx_queue[-1] == diagnostic_request_to_address
+
+
 def test_receive_diagnostic(mock_socket):
     sut = DoIPClient(test_ip, test_logical_address)
     mock_socket.rx_queue.append(diagnostic_result)
@@ -526,10 +547,6 @@ def test_request_vehicle_identification_with_ein(mock_socket):
     assert mock_socket.tx_queue[-1] == vehicle_identification_request_with_ein
     assert result.vin == "1" * 17
     assert result.logical_address == 0x1234
-    assert result.eid == b"1" * 6
-    assert result.gid == b"2" * 6
-    assert result.further_action_required == 0x00
-    assert result.vin_sync_status == 0x00
 
 
 def test_request_vehicle_identification_with_vin(mock_socket):
@@ -545,6 +562,42 @@ def test_request_vehicle_identification_with_vin(mock_socket):
     assert result.vin_sync_status == 0x00
 
 
+def test_get_entity(mock_socket):
+    mock_socket.rx_queue.append(vehicle_identification_response)
+    _, result = DoIPClient.get_entity()
+    assert mock_socket.tx_queue[-1] == vehicle_identification_request
+    assert result.vin == "1" * 17
+    assert result.logical_address == 0x1234
+    assert result.eid == b"1" * 6
+    assert result.gid == b"2" * 6
+    assert result.further_action_required == 0x00
+    assert result.vin_sync_status == 0x00
+
+
+def test_get_entity_with_ein(mock_socket):
+    mock_socket.rx_queue.append(vehicle_identification_response)
+    _, result = DoIPClient.get_entity(eid=b"1" * 6)
+    assert mock_socket.tx_queue[-1] == vehicle_identification_request_with_ein
+    assert result.vin == "1" * 17
+    assert result.logical_address == 0x1234
+    assert result.eid == b"1" * 6
+    assert result.gid == b"2" * 6
+    assert result.further_action_required == 0x00
+    assert result.vin_sync_status == 0x00
+
+
+def test_get_entity_with_vin(mock_socket):
+    mock_socket.rx_queue.append(vehicle_identification_response)
+    _, result = DoIPClient.get_entity(vin="1" * 17)
+    assert mock_socket.tx_queue[-1] == vehicle_identification_request_with_vin
+    assert result.vin == "1" * 17
+    assert result.logical_address == 0x1234
+    assert result.eid == b"1" * 6
+    assert result.gid == b"2" * 6
+    assert result.further_action_required == 0x00
+    assert result.vin_sync_status == 0x00
+
+
 def test_request_diagnostic_power_mode(mock_socket):
     sut = DoIPClient(test_ip, test_logical_address)
     mock_socket.rx_queue.append(diagnostic_power_mode_response)
@@ -616,11 +669,13 @@ def test_await_ipv6(mock_socket):
     except TimeoutError:
         pass
     assert mock_socket._network == socket.AF_INET6
-    assert mock_socket._bound_ip == 'ff02::1'
+    assert mock_socket._bound_ip == "ff02::1"
     assert mock_socket._bound_port == 13400
     assert mock_socket.opts == {
         socket.SOL_SOCKET: {socket.SO_REUSEADDR: True},
-        IPPROTO_IPV6: {socket.IPV6_JOIN_GROUP: b'\xff\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x01\x00\x00\x00\x00'},
+        IPPROTO_IPV6: {
+            socket.IPV6_JOIN_GROUP: b"\xff\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x01\x00\x00\x00\x00"
+        },
     }
 
 
@@ -637,4 +692,132 @@ def test_await_ipv4(mock_socket):
     assert mock_socket._bound_port == 13400
     assert mock_socket.opts == {
         socket.SOL_SOCKET: {socket.SO_REUSEADDR: True, socket.SO_BROADCAST: True},
-    }
\ No newline at end of file
+    }
+
+
+def test_exception_from_blocking_ssl_socket(mock_socket, mocker):
+    """SSL sockets behave slightly different than regular sockets in
+    non-blocking mode. They won't raise BlockingIOError but SSLWantWriteError
+    or SSLWantReadError instead.
+
+    See: https://docs.python.org/3/library/ssl.html#notes-on-non-blocking-sockets
+    """
+    sut = DoIPClient(test_ip, test_logical_address)
+
+    try:
+        sut._tcp_sock.recv = mocker.Mock(side_effect=ssl.SSLWantReadError)
+        sut._tcp_socket_check()
+        sut._tcp_sock.recv = mocker.Mock(side_effect=ssl.SSLWantWriteError)
+        sut._tcp_socket_check()
+    except (ssl.SSLWantReadError, ssl.SSLWantWriteError) as exc:
+        pytest.fail(f"Should not raise exception: {exc.__class__.__name__}")
+
+
+def test_use_secure_uses_default_ssl_context(mock_socket, mocker):
+    """Wrap socket with default SSL-context when use_secure=True"""
+    mocked_default_context = mocker.patch.object(
+        ssl, "create_default_context", autospec=True
+    )
+    sut = DoIPClient(
+        test_ip, test_logical_address, use_secure=True, activation_type=None
+    )
+    mocked_default_wrap_socket = mocked_default_context.return_value.wrap_socket
+    mocked_default_wrap_socket.assert_called_once_with(mock_socket)
+
+
+def test_use_secure_with_external_ssl_context(mock_socket, mocker):
+    """Wrap socket with user provided SSL-context when use_secure=ssl_context"""
+    original_context = ssl.SSLContext
+    mocked_external_context = mocker.patch.object(ssl, "SSLContext", autospec=True)
+    mocked_default_context = mocker.patch.object(
+        ssl, "create_default_context", autospec=True
+    )
+
+    # Unmock the SSLContext
+    ssl.SSLContext = original_context
+
+    sut = DoIPClient(
+        test_ip,
+        test_logical_address,
+        use_secure=mocked_external_context,
+        activation_type=None,
+    )
+
+    mocked_default_wrap_socket = mocked_default_context.return_value.wrap_socket
+    assert (
+        not mocked_default_wrap_socket.called
+    ), "Socket should *not* get wrapped using default context."
+
+    mocked_external_wrap_socket = mocked_external_context.wrap_socket
+    mocked_external_wrap_socket.assert_called_once_with(mock_socket)
+
+
+@pytest.mark.parametrize(
+    "vm_specific, exc",
+    (
+        (None, False),
+        (0, False),
+        (-1, True),
+        (0xFFFFFFFF, False),
+        (0x100000000, True),
+        ("0x1", True),
+        (10.0, True),
+    ),
+)
+def test_vm_specific_setter(mock_socket, mocker, vm_specific, exc):
+    sut = DoIPClient(test_ip, test_logical_address, auto_reconnect_tcp=True)
+    if exc:
+        with pytest.raises(ValueError):
+            sut.vm_specific = vm_specific
+    else:
+        sut.vm_specific = vm_specific
+        assert sut.vm_specific == vm_specific
+
+
+def test_vm_specific_static_value(mock_socket, mocker):
+    request_activation_spy = mocker.spy(DoIPClient, "request_activation")
+    mock_socket.rx_queue.append(successful_activation_response_with_vm)
+
+    sut = DoIPClient(
+        test_ip,
+        test_logical_address,
+        auto_reconnect_tcp=True,
+        activation_type=None,
+        vm_specific=0x01020304,
+    )
+    sut.request_activation(
+        activation_type=RoutingActivationRequest.ActivationType.Default
+    )
+    assert mock_socket.tx_queue[-1] == activation_request_with_vm
+    assert request_activation_spy.call_count == 1
+
+
+def test_vm_specific_request_activation_bad_value(mock_socket, mocker):
+    request_activation_spy = mocker.spy(DoIPClient, "request_activation")
+    mock_socket.rx_queue.append(successful_activation_response_with_vm)
+
+    sut = DoIPClient(
+        test_ip,
+        test_logical_address,
+        auto_reconnect_tcp=True,
+        activation_type=None,
+        vm_specific=0x01020304,
+    )
+    with pytest.raises(ValueError):
+        sut.request_activation(
+            activation_type=RoutingActivationRequest.ActivationType.Default,
+            vm_specific=-1,
+        )
+
+
+def test_vm_specific_verification_in_init(mock_socket, mocker):
+    request_activation_spy = mocker.spy(DoIPClient, "request_activation")
+    mock_socket.rx_queue.append(successful_activation_response_with_vm)
+    with pytest.raises(ValueError):
+        sut = DoIPClient(
+            test_ip,
+            test_logical_address,
+            auto_reconnect_tcp=True,
+            activation_type=None,
+            vm_specific=-1,
+        )
